/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "omc_typ.h"
#include "lha_typ.h"
#include "hierview.h"
#include "c2h_typ.h"
#include "bpps_usage.h"
#include "cmc_typ.h"
#include "rpm_typ.h"
#include "hsc_typ.h"

static void PrintError()
{
	// PrintLicenseStatement("bpps v1.0.8"); 
	fprintf(stderr,"%s",BPPS_VERSION); print_error(USAGE_BPPS); 
}

int     main(int argc, char *argv[])
{
	int	i,a,A,Argc,x;
	Int4	rtn=0;
	double	d;
	char	**Argv,mode,str[200],c; 
	TurnOffLicenseStatement();
	if(argc < 2) PrintError(); else mode=argv[1][0];
	if(strchr("A123EQH",mode)==NULL) PrintError();
	if(argv[1][1] != 0) PrintError();
	NEWP(Argv,argc+8,char);
	Argv[0]=argv[0]; Argc=argc-1;	// skip over argv[1]...
	if(mode != 'H'); for(a=2,A=1; a < argc; a++,A++){ Argv[A]=argv[a]; }
	switch (mode){
	   case 'A':
	     {	
		char pdb[100];
		if(argc < 3) print_error(USAGE_BPPS);
		if(Argv[2] && Argv[2][0] != '-'){ sprintf(pdb,"-pdb=%s",Argv[2]); Argv[3]=pdb; }
		else Argv[3]=0;
		Int4 NumSets=0; Argc=2;
	        { 
		  omc_typ omc(1,Argc,Argv); rtn=omc.Run(); omc.PrintTime(stderr); 
		  hpt_typ *Hpt=omc.RtnHpt(); NumSets=Hpt->NumSets()-1; 
		} run_hieraln(2,Argc,Argv); 
		if(Argv[3]) Argc=4; else Argc=3;
		sprintf(str,"%s_himsa",Argv[1]); Argv[1]=AllocString(str);
		for(i=2; i <= NumSets; i++){
		    sprintf(str,"%d",i); Argv[2]=str; run_hierview(3,Argc,Argv); 
		} free(Argv[1]);
	     } break;
	   case 'Q': {
	      if(argc < 4){ omc_typ  omc('Q',Argc,Argv); } 
	      if(!isdigit(argv[3][0])) print_error("FATAL: The 3rd argument must be a positive integer.");
	      unsigned long X=strtoul(argv[3],(char **) NULL,10);
	      if(X < 1) print_error("FATAL: The 3rd argument must be a positive integer.");
	      // fprintf(stderr,"X=%u\n",X);
	      // if(X == ULONG_MAX) print_error(USAGE_BPPS);
	      if(X == ULONG_MAX) { omc_typ  omc('Q',Argc,Argv); }  // this triggers usage...
	      sprintf(str,"%s.chk",argv[2]);
	      FILE *fp=fopen(str,"r");
	      if(fp != NULL) print_error("FATAL: checkpoint (*.chk) output file present\n");
	      sprintf(str,"-Focus=1:%u",X); Argv[2]=AllocString(str);
              for(Int4 arg=3; arg < argc; arg++){
                if(argv[arg][0] != '-') break;	// this will be reported later.
                switch(argv[arg][1]) {
                 case 'm': 
		  if(sscanf(argv[arg],"-minnats=%lf",&d)==1) continue;
		  if(sscanf(argv[arg],"-minsize=%d",&x)==1) continue;
		  if(sscanf(argv[arg],"-maxdepth=%d",&x) == 1) continue;
		  if(sscanf(argv[arg],"-maxcol=%d",&x) == 1) continue;
		  if(sscanf(argv[arg],"-mtf=%s",&str) == 1) continue;
		  else argv[arg][1] = '#';	// triggers print usage for Q
		  break;
                case 'r': 
		  if(sscanf(argv[arg],"-rtfF=%d",&x)==1) continue;
		  if(sscanf(argv[arg],"-rtfP=%c",&c)==1) continue;
		  else argv[arg][1] = '#';	// triggers print usage for Q
		  break;
                case 's': 
		  if(sscanf(argv[arg],"-seed=%d",&x)!=1) continue;
		  else argv[arg][1] = '#';	// triggers print usage for Q
		  break;
                case 'u': 
		  if(strcmp("-usr",argv[arg]) == 0) continue; 
		  else argv[arg][1] = '#';	// triggers print usage for Q
		  break;
                default : argv[arg][1] = '#';	// triggers print usage for Q
               }
	      } omc_typ  omc('Q',Argc,Argv); rtn=omc.Run(); free(Argv[2]);
	    } break;
	   case '1': { omc_typ	omc(1,Argc,Argv); rtn=omc.Run(); } break;
	   case '2': { run_hieraln(2,Argc,Argv); } break;
	   case '3': { run_hierview(3,Argc,Argv); } break;
	   case 'H': { 
		char *pdb_list=0;
		BooLean heatmap=FALSE;
		for(a=2,A=1; a < argc; a++){ 
		   BooLean skip=FALSE;
fprintf(stderr,"'%s'\n",argv[a]);
                   if(argv[a][0] == '-'){
                     switch(argv[a][1]) {
		       case 'p':	// pdb_list
		        if(sscanf(argv[a],"-pdb_list=%s",str) == 1){
			   pdb_list=AllocString(str); skip=TRUE;
			} break;
		       case 'h':	// heatmaps
		        if(strcmp("-heatmap",argv[a]) == 0){
			  if(argv[a][8]==0){
			     heatmap=TRUE; skip=TRUE;
fprintf(stderr,"%d(skip): '%c'\n",a,argv[a][8]);
			  }
else fprintf(stderr,"%d: '%c'\n",a,argv[a][8]);
			} break;
		     }
		   }
		   if(!skip){ Argv[A]=argv[a]; A++; }
		} Argc=A;
		for(i=0;i < Argc; i++) fprintf(stderr,"'%s' ",Argv[i]);
		fprintf(stderr,"\n");
		RuncmcBPPS(Argc,Argv); 
fprintf(stderr,"!!!!!!!!!!!!!! ==> DEBUG 1\n");
		if(pdb_list){
		  Argv[0] = AllocString("sarp");		
		  Argv[1] = AllocString(pdb_list); 
		  sprintf(str,"%s",argv[2]);	
		  Argv[2] = AllocString(str); Argc=3;
		  { hsc_typ hsc(Argc, Argv); hsc.Run( ); }
		  for(i=0;i < Argc; i++) free(Argv[i]);
		}
fprintf(stderr,"!!!!!!!!!!!!!! ==> DEBUG 2\n");
		if(heatmap){
fprintf(stderr,"!!!!!!!!!!!!!! ==> DEBUG 3\n");
		  Argv[0] = AllocString("pttrn_mtch");		
		  sprintf(str,"%s_new.mma",argv[2]);	
		  Argv[1] = AllocString(str);		
		  sprintf(str,"%s_new.hpt",argv[2]);	
		  Argv[2] = AllocString(str); Argc=3;
		  rpm_typ *rpm = new rpm_typ(Argc,Argv);
        	  rpm->PutHeatMaps(argv[2]); delete rpm;
		  for(i=0;i < Argc; i++) free(Argv[i]);
		}
		     } break;
	   case 'E': { c2h_typ c2h(mode,Argc,Argv); c2h.Run(); } break;
	   case 'U': { print_error("bpps U option not yet implemented"); } break;
	   default: print_error(USAGE_BPPS); break;
	} free(Argv); return 0;
}

