/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "cmc_typ.h"

void    cmc_typ::FWriteSST(FILE *fp)
{
	Int4 N=Hpt->NumBPPS();
	Int4 rtn=fwrite(&N,sizeof(Int4),1,fp);
       	for(Int4 i=1; i<=Hpt->NumBPPS(); i++){
	    bpps_typ *bpps=che[i]->BPPS();
	    sst_typ *tmpsst=bpps->RtnSST();
	    Int4 k=bpps->LenPattern( );
	    assert(k == LenSeqCMSA(1,MainCMA));
	    rtn=fwrite(&k,sizeof(Int4),1,fp);
       	    rtn=fwrite(tmpsst,sizeof(sst_typ),k+2,fp);      // made in cmc_typ
       	    if(rtn != k+2) print_error("WriteSST() input error 4");
       	}
}

sst_typ	**cmc_typ::FReadSST(FILE *fp,Int4 &numBPPS,Int4 &len)
{
	Int4 K,k,rtn=0;
	rtn=fread(&numBPPS,sizeof(Int4),1,fp);
	sst_typ **tmpsst; NEWP(tmpsst, numBPPS+3,sst_typ);
       	for(Int4 i=1; i<=numBPPS; i++){
	    rtn=fread(&k,sizeof(Int4),1,fp);
	    NEW(tmpsst[i], k+3,sst_typ);
	    if(i==1) K=k; else assert(K==k);
       	    rtn=fread(tmpsst[i],sizeof(sst_typ),k+2,fp);      // made in cmc_typ
       	    if(rtn != k+2) print_error("ReadSST() input error 4");
       	} len=K;
	return tmpsst;
}

void	cmc_typ::ReadCheckpoint(char *outfilename)
{
	NEW(checkpoint,1,cmc_checkpt_type);
	// read hpt.
	if(outfilename==0) outfilename=infile;

#if 0
        fp=open_file(name,".chk","w"); 
        assert(hpt->NumSMA != 0);
          WriteSets(fp,hpt->NumSets,info->set);
          fprintf(fp,"seed=%d; N=%d.\n",seed,NumSMA);
          for(Int4 i=1; i <= NumSMA; i++){ PutCMSA(fp,sma[i]); }
          hpt->Put(fp,TRUE,FALSE,TRUE);
#else
        FILE *fp = open_file(outfilename,"_chk.cma","r");
	// Read MainCMA file.
	checkpoint->main_cma=ReadCMSA(fp,AB); fclose(fp);

	fp = open_file(outfilename,"_chk.hpt","r");
	checkpoint->hpt= new hpt_typ(fp); fclose(fp);
	// Hpt->Put(fp); 

	// open binary file.
              	fp = open_file(outfilename,"_chk.cmc","r");

	// read sequence weights.
              	checkpoint->hsw=FReadHSW(fp,AB,checkpoint->main_cma); 

	// read elementary sets.
        checkpoint->set=ReadSets(fp,checkpoint->num_sets); 
	checkpoint->labeled=checkpoint->set[checkpoint->num_sets];
	checkpoint->set[checkpoint->num_sets]=0;
	checkpoint->num_sets--;

	// read patterns.
	Int4 k;
	checkpoint->sst=FReadSST(fp,checkpoint->num_ssts,k); 
        assert(k == LenSeqCMSA(1,checkpoint->main_cma)); fclose(fp);
#endif

#if 1   // if (NumSeedAln + 1) != num_passed_in_sets then delete empty sets.
        // This needs to be done properly; TypeOfSet[i] corresponds to seed alignments, not sets.
        Int4 npis=0, num_ssts=0;
        set_typ *pis; NEW(pis,checkpoint->num_sets+3, set_typ);
	sst_typ	**sst; NEWP(sst,checkpoint->num_ssts+3,sst_typ);
        if((checkpoint->hpt->NumSets()+ 1) != checkpoint->num_sets || TRUE){
           for(Int4 i=1; i <= checkpoint->num_sets; i++){
                if(i <= checkpoint->hpt->NumSets() && checkpoint->hpt->TypeOfSet(i) != '?' && CardSet(checkpoint->set[i]) ==0){
			fprintf(stderr,"Removing empty set %d\n",i);
                        NilSet(checkpoint->set[i]); free(checkpoint->sst[i]);
                } else {
			npis++; pis[npis]=checkpoint->set[i]; 
			// if(i <= checkpoint->num_ssts) 
			sst[npis]=checkpoint->sst[i]; 
		}
           } free(checkpoint->set); checkpoint->set=pis; checkpoint->num_sets=npis; 
		checkpoint->sst= sst;  checkpoint->num_ssts=npis;
        }
#endif
}

BooLean	cmc_typ::WriteCheckpoint(char *outfilename)
{
	Int4	i,j,NumSMA;
	char	str[500];
        FILE *fp=0;
	if(outfilename==0) outfilename=infile;
	if(Hpt->NumSets() < 3) return FALSE; // Don't print if only root node.

#if 1   // *.chk single output file.
        fp=open_file(outfilename,".chk","w");
	set_typ *set=this->CopyOfSeqSets();
	{
          bpcp_typ bpcp(RandomSeed,Hpt->NumSets(),set,NumDisplayCMA,DisplayCMA,Hpt);
          bpcp.Write(fp); 
	} fclose(fp); fp=0;
        for(i=1; i <= Hpt->NumSets(); i++){ if(set[i]) NilSet(set[i]); } free(set);
        return TRUE;
#else	// old format...
	//============== Save sequence sets ==========
	fp = open_file(outfilename,"_chk.sets","w");
	set_typ *set=this->CopyOfSeqSets();
        WriteSets(fp,Hpt->NumSets(),set); fclose(fp);

	cma_typ	*sma=0;
	cma_typ *sma=DisplayCMA;
#if 0
	//============== Save seed sequences for each group ==========
	NumSMA=Hpt->NumSets();
	if(passed_in_sma == 0){
	   NEW(sma,Hpt->NumSets() +3,cma_typ);
	   for(j=1; j < Hpt->NumSets(); j++){
		sprintf(str,"%s",Hpt->ElmntSetName(j));
		sma[j]=this->RtnCsqAsCMSA(j,str);  // won't redefine xsst[i].
	   }
	} else sma=passed_in_sma;
	fp = open_file(outfilename,"_chk.sma","w");
        fprintf(fp,"seed=%d; N=%d.\n",RandomSeed,NumSMA);
        for(i=1; i <= NumSMA; i++){ PutCMSA(fp,sma[i]); } fclose(fp);
	if(sma==passed_in_sma) sma=0;
	
	//============= save sequence sets: =============
	Int4 g; assert(DidRestoreBest);
	set_typ *Set=0; NEW(Set,Hpt->NumSets()+3,set_typ);
	for(g=1; g<= Hpt->NumSets(); g++){ Set[g]=BestSet[g]; } 
	Set[g]=Labeled; WriteSets(fp,g,Set);  free(Set);
#endif

	//============= save Hpt with patterns: =============
	fp = open_file(outfilename,"_chk.hpt","w");
#if 0
	Hpt->Put(fp,FALSE,FALSE,FALSE);  // put settings; retain internal; don't putArg. 
	fprintf(fp,"\nSettings:\n");
	for(Int4 g=1; g<= Hpt->NumBPPS(); g++){ 
	   if(Hpt->nArg(g) == 0){
	      fprintf(fp,"%d.Group%d -P=",g,g);
	      che[g]->PutBestPatterns(fp,FALSE);
	      // ^Use column not actual sequence position.
	   } else {
	     fprintf(fp,"%d.%s",g,Hpt->GrpName(g));
	     char **Argv=Hpt->Argv(g);
	     for(Int4 i=0; i < Hpt->nArg(g); i++){
		if(strncmp(Argv[i],"-P",2) != 0 && strncmp(Argv[i],"- ",2) != 0){
			fprintf(fp," %s",Argv[i]); 
		}
	     }
	     // fprintf(fp,"%d.%s -P=",g,Hpt->GrpName(g));
	     fprintf(fp," -P=");
	     che[g]->PutBestPatterns(fp,FALSE);
	     // use column not actual sequence position.
	   } // fprintf(fp,"\n");  // PutBestPatterns() ends with rtn char.
	} fprintf(fp,"\n\n"); 
#else
	this->PutHptPttrns(fp);
#endif
	fclose(fp);


	//==================== Free Memory =================
        if(sma){
                for(i=1; i < Hpt->NumSets(); i++){
                   if(sma[i]) TotalNilCMSA(sma[i]); sma[i] = 0;
                } free(sma);
        }
	if(set){
                for(i=1; i <= Hpt->NumSets(); i++){
                   if(set[i]) NilSet(set[i]); set[i] = 0;
                } free(set);
        } return TRUE;
#endif
}

#if 0	// old version that doesn't work.
void	cmc_typ::WriteCheckpoint(char *outfilename)
{
	if(outfilename==0) outfilename=infile;
#if 0
        fp = open_file(argv[1],".hsw","w");
        CreateWriteHSW(fp,MainCMA); fclose(fp);
#endif
#if 0
	// first go through and find failed columns and rows and remove
	// check for IsFailedBPPS[] & IsFailedSet[];
#endif
	// open hpt file:
        FILE *fp;
	fp = open_file(outfilename,"_chk.hpt","w");
#if 0
	Hpt->Put(fp); fclose(fp);
#else
	// Hpt->Put(fp,FALSE);  // don't put settings...
	// Hpt->Put(fp,TRUE,FALSE,TRUE);  // put settings; don't ignore internal; putArg. 
	// Hpt->Put(fp,TRUE,FALSE,FALSE);  // put settings; retain internal; don't putArg. 
	Hpt->Put(fp,FALSE,FALSE,FALSE);  // put settings; retain internal; don't putArg. 
	fprintf(fp,"\nSettings:\n");
	for(Int4 g=1; g<= Hpt->NumBPPS(); g++){ 
	   if(Hpt->nArg(g) == 0){
	      fprintf(fp,"%d.Group%d -P=",g,g);
	      che[g]->PutBestPatterns(fp,FALSE);
	      // ^Use column not actual sequence position.
	   } else {
	     fprintf(fp,"%d.%s",g,Hpt->GrpName(g));
	     char **Argv=Hpt->Argv(g);
	     for(Int4 i=0; i < Hpt->nArg(g); i++){
		if(strncmp(Argv[i],"-P",2) != 0 && strncmp(Argv[i],"- ",2) != 0){
			fprintf(fp," %s",Argv[i]); 
		}
	     }
	     // fprintf(fp,"%d.%s -P=",g,Hpt->GrpName(g));
	     fprintf(fp," -P=");
	     che[g]->PutBestPatterns(fp,FALSE);
	     // use column not actual sequence position.
	   } /// fprintf(fp,"\n");  // PutBestPatterns() ends with rtn char.
	} fprintf(fp,"\n\n"); fclose(fp);
#endif

        fp = open_file(outfilename,"_chk.cma","w");
	PutCMSA(fp,MainCMA); fclose(fp);

	// open binary file:
        fp = open_file(outfilename,"_chk.cmc","w");

	// write sequence weights...
	if(passed_in_hsw) FWriteHSW(fp,passed_in_hsw); 
	else if(checkpoint && checkpoint->hsw) FWriteHSW(fp,checkpoint->hsw); 
	else {
	   hsw_typ hsw=chn[1]->RtnHSW(1);	// created in cmc_typ.cc
	   FWriteHSW(fp,hsw); // NilHSW(hsw);  this is owned by chn[1]->swt; don't delete it!
	}

	// write sets...
	Int4 g; assert(DidRestoreBest);
	set_typ *Set=0; NEW(Set,Hpt->NumSets()+3,set_typ);
	for(g=1; g<= Hpt->NumSets(); g++){ Set[g]=BestSet[g]; } 
	Set[g]=Labeled; WriteSets(fp,g,Set);  free(Set);

	// write patterns.
	FWriteSST(fp); fclose(fp);
}

Int4	cmc_typ::PutCheckPoint(char *string,BooLean change)
// main routine to perform operations on cmc_typ cmc.
{
	if(Hpt->NumSets() < 3) return 0; // Don't print if only root node.
	char	*name=AllocString(string);
char str[500];
	// this->RestoreVeryBest( );  // afn: 4/28/2022
	this->RestoreBest( );  // afn: 5/10/2022
	Int4	*P=0,id,sq,i,j,pI,I,J,r,R,m,n,p,x,cID=0;
	info_typ	*info = new info_typ; info->Init(MaxNumNodesPlus); 
	hpt_typ	*hpt=0;
	sst_typ **xsst=this->RtnCopyOfSSTs();
	char    **pttrn; NEWP(pttrn,MaxNumNodesPlus+3,char);

#if 0
	set_typ *set=this->CopyOfBestTreeSets();
#else
	set_typ *set=this->CopyOfSeqSets();
#endif
	hpt=Hpt->Copy();  info->hpt=hpt;  info->T=0; 
// fprintf(stderr,"taking checkpoint\n");
	for(j=1; j < hpt->NumSets(); j++){
		if(change) {
		   id=Hpt->ItoSetID(j); sprintf(str,"Set%d",id);
		} else { sprintf(str,"%s",Hpt->ElmntSetName(j)); }
		info->sma[j]=this->RtnCsqAsCMSA(j,str);  // won't redefine xsst[i].
		pttrn[j]=SST2ArgStrAlpha(xsst[j],this->RtnLengthMainCMSA(),AB);
		info->set[j]=set[j];
	} info->set[hpt->NumSets()] = set[j]; // info->set=set; set=0;
	for(i=1; i <= Hpt->NumBPPS(); i++){ if(xsst[i]) free(xsst[i]); } free(xsst); 
     	for(i=1; i <= hpt->NumBPPS(); i++){
// fprintf(stderr,"pttrn[%d]=%s\n",i,pttrn[i]);
            if(pttrn[i]){
		char *tmp[3]; tmp[0]=pttrn[i]; hpt->ReSetArgv(i,1,tmp); free(pttrn[i]); 
		if(change){
		   id=hpt->ItoSetID(i); sprintf(str,"Set%d",id); hpt->ReNameGroup(i,str);
		} 
	    }
     	} 
	// hpt->Put(stderr,TRUE,FALSE,TRUE);
fprintf(stderr,"DEBUG: file name = %s\n",name);

        FILE *fp=0;
#if 0	// *.chk single output file.
        fp=open_file(name,".chk","w"); 
	{
	  bpcp_typ bpcp(RandomSeed,hpt->NumSets(),info->set,hpt->NumSets()-1,
				info->sma,hpt);
	  bpcp.Write(fp);
	} fclose(fp); fp=0; 
#elif 1
        fp=open_file(name,".chk","w");
	Int4 NumSMA;
	for(i=1, NumSMA=0; info->sma[i]; i++) NumSMA++;
        assert(NumSMA != 0);
        WriteSets(fp,hpt->NumSets(),info->set);
        fprintf(fp,"seed=%d; N=%d.\n",RandomSeed,NumSMA);
        for(Int4 i=1; i <= NumSMA; i++){ PutCMSA(fp,info->sma[i]); }
        // hpt->Put(fp,TRUE,FALSE,TRUE);
        hpt->Put(fp,TRUE,FALSE,FALSE);
#else	// output old format files.
	fp=open_file(name,".hpt","w"); hpt->Put(fp,TRUE,FALSE,TRUE); fclose(fp); fp=0; 
        fp=open_file(name,".sets","w"); WriteSets(fp,hpt->NumSets(),info->set);
	fclose(fp); fp=0; 
        fp=open_file(name,".sma","w");
	for(i=1; i < info->hpt->NumSets(); i++){ PutCMSA(fp,info->sma[i]); }
	fclose(fp); fp=0; 
        fp=open_file(name,".seed","w"); fprintf(fp,"-seed=%d\n",RandomSeed); fclose(fp); fp=0;
#endif
	if(info->hpt){ 
	  if(info->sma){
		for(i=1; i < info->hpt->NumSets(); i++){
		   if(info->sma[i]) TotalNilCMSA(info->sma[i]); info->sma[i] = 0;
		} 
	  } if(info->set){
		for(i=1; i <= info->hpt->NumSets(); i++){
		   if(info->set[i]) NilSet(info->set[i]); info->set[i] = 0;
		}
	  } delete info->hpt; 
	} info->Free(); delete info; 
	if(set) free(set); free(pttrn); if(P) free(P); free(name);
	return 1;
}

#endif

